@@ -7,8 +7,8 @@ r"""
                                    /_/
 *******************************************************************************
 
-Metaprogramming and mixin tools that implement property tuple support for
-python enumeration classes.
+Metaprogramming and mixin tools that implement property tuple and method
+specialization support for python enumeration classes.
 
 .. todo::
     Given how dynamic the typing is in this module, static type checking is
@@ -31,7 +31,7 @@ except ImportError:  # pragma: no cover
     cached_property = property  # pylint: disable=C0103
 
 
-VERSION = (1, 3, 3)
+VERSION = (1, 5, 2)
 
 __title__ = 'Enum Properties'
 __version__ = '.'.join(str(i) for i in VERSION)
@@ -48,6 +48,7 @@ __all__ = [
     'EnumPropertiesMeta',
     'SymmetricMixin',
     'DecomposeMixin',
+    'specialize',
     'p',
     's'
 ]
@@ -76,7 +77,8 @@ class _SProp(_Prop):
 
 def s(  # pylint: disable=C0103
     prop_name,
-    case_fold=False
+    case_fold=False,
+    match_none=False
 ):
     """
     Add a symmetric property. Enumeration values will be coercible from this
@@ -85,12 +87,15 @@ def s(  # pylint: disable=C0103
     :param prop_name: The name of the property
     :param case_fold: If False, symmetric lookup will be
         case sensitive (default)
+    :param match_none: If True, none values will be symmetric, if False
+        (default), none values for symmetric properties will not map back to
+        the enumeration value.
     :return: a named symmetric property class
     """
     return type(
         prop_name,
         (_SProp,),
-        {'symmetric': True, 'case_fold': case_fold}
+        {'symmetric': True, 'case_fold': case_fold, 'match_none': match_none}
     )
 
 
@@ -108,6 +113,33 @@ def p(prop_name):  # pylint: disable=C0103
     return type(prop_name, (_Prop,), {'symmetric': False})
 
 
+class _Specialized:  # pylint: disable=R0903
+    """
+    A member specialization identifier class - private. Used to wrap
+    specialized member functions and tag them with their enum values so the
+    Enum classdict can identify them.
+
+    :param wrapped: The wrapped member function
+    :param value: The value to specialize for
+    """
+    def __init__(self, wrapped, values):
+        self.wrapped = wrapped
+        self.values = values
+
+
+def specialize(*values):
+    """
+    A decorator to specialize a method for a given enumeration value.
+
+    :param values: The enumeration value(s) to specialize
+    :return: A decorated specialized member method
+    """
+    def specialize_decorator(method):
+        return _Specialized(method, values)
+
+    return specialize_decorator
+
+
 class SymmetricMixin:  # pylint: disable=R0903
     """
     This mixin enables symmetric Enum_ creation from properties marked
@@ -186,25 +218,26 @@ class SymmetricMixin:  # pylint: disable=R0903
             except KeyError:
                 pass
 
-        for coerce_to in cls._ep_coerce_types_:
-            try:
-                val = coerce_to(value)
+        if value is not None:
+            for coerce_to in cls._ep_coerce_types_:
                 try:
-                    return cls._value2member_map_[val]
-                except KeyError:
+                    val = coerce_to(value)
+                    try:
+                        return cls._value2member_map_[val]
+                    except KeyError:
+                        pass
+
+                    try:
+                        return cls._ep_symmetric_map_[val]
+                    except KeyError:
+                        pass
+
+                    if isinstance(val, str):
+                        return cls._ep_isymmetric_map_[_do_casenorm(val)]
+
+                except Exception:  # pylint: disable=W0703
                     pass
 
-                try:
-                    return cls._ep_symmetric_map_[val]
-                except KeyError:
-                    pass
-
-                if isinstance(val, str):
-                    return cls._ep_isymmetric_map_[_do_casenorm(val)]
-
-            except (KeyError, TypeError, ValueError):
-                pass
-
         return super()._missing_(value)
 
 
@@ -249,14 +282,14 @@ class EnumPropertiesMeta(enum.EnumMeta):
 
     # members reserved for use by EnumProperties
     RESERVED = [
-        'enum_properties',
+        '_properties_',
         '_ep_coerce_types_',
         '_ep_symmetric_map_',
         '_ep_isymmetric_map_'
     ]
 
     @classmethod
-    def __prepare__(mcs, cls, bases):  # pylint: disable=W0221
+    def __prepare__(mcs, cls, bases, **kwargs):  # pylint: disable=W0221,R0915
         """
         Strip properties out of inheritance and record them on our class
         dictionary for per-value based assignment in ``__new__``.
@@ -282,7 +315,7 @@ class EnumPropertiesMeta(enum.EnumMeta):
             else:
                 real_bases.append(base)
 
-        class_dict = super().__prepare__(cls, tuple(real_bases))
+        class_dict = super().__prepare__(cls, tuple(real_bases), **kwargs)
 
         class _PropertyEnumDict(class_dict.__class__):
             """
@@ -295,6 +328,7 @@ class EnumPropertiesMeta(enum.EnumMeta):
 
             _ep_properties_ = properties
             _ep_symmetric_properties = symmetric_properties
+            _specialized_ = {}
 
             def __init__(self):
                 super().__init__()
@@ -304,7 +338,10 @@ class EnumPropertiesMeta(enum.EnumMeta):
                     self[item] = value
 
             def __setitem__(self, key, value):
-                if key in EnumPropertiesMeta.EXPECTED:
+                if isinstance(value, _Specialized):
+                    for en_val in value.values:
+                        self._specialized_.setdefault(en_val, {})[key] = value
+                elif key in EnumPropertiesMeta.EXPECTED:
                     dict.__setitem__(self, key, value)
                 elif key in EnumPropertiesMeta.RESERVED:
                     raise ValueError(f'{key} is reserved.')
@@ -350,7 +387,7 @@ class EnumPropertiesMeta(enum.EnumMeta):
                             ) from type_err
 
                     elif key in class_dict._member_names:
-                        remove = True
+                        remove = True  # pragma: no cover
 
                     super().__setitem__(key, value)
 
@@ -375,16 +412,18 @@ class EnumPropertiesMeta(enum.EnumMeta):
             mcs,
             classname,
             bases,
-            classdict
+            classdict,
+            **kwargs
     ):
         """
         Enumeration class construction runs in the following stages:
 
         1) pass up the inheritance tree to build the initial enumeration class.
-        2) Add property value to enumeration value maps for each property and
+        2) Add method specializations to each class
+        3) Add property value to enumeration value maps for each property and
             the property accessors that use them
-        3) Add casefolded symmetric maps for any symmetric properties
-        4) Add any symmetric builtin properties to our symmetric maps
+        4) Add casefolded symmetric maps for any symmetric properties
+        5) Add any symmetric builtin properties to our symmetric maps
 
         :raises ValueError: if ``_symmetric_builtins_`` is specified
             incorrectly, or if non-hashable values are provided for a
@@ -394,14 +433,28 @@ class EnumPropertiesMeta(enum.EnumMeta):
             mcs,
             classname,
             tuple(base for base in bases if not issubclass(base, _Prop)),
-            classdict
+            classdict,
+            **kwargs
         )
         cls._ep_coerce_types_ = []
         cls._ep_symmetric_map_ = cls._member_map_
         cls._ep_isymmetric_map_ = {}
-        cls.enum_properties = list(classdict._ep_properties_.keys())
+        cls._properties_ = list(classdict._ep_properties_.keys())
+
+        for val in cls:
+            for member_name, specialization in classdict._specialized_.get(
+                    val.value, {}
+            ).items():
+                # use descriptor binding
+                setattr(
+                    val,
+                    member_name,
+                    specialization.wrapped.__get__(val)
+                )
 
         def add_sym_lookup(prop, p_val, enum_inst):
+            if p_val is None and not prop.match_none:
+                return
             if not isinstance(p_val, Hashable):
                 raise ValueError(
                     f'{cls}.{prop}:{p_val} is not hashable. Symmetrical '
@@ -422,25 +475,18 @@ class EnumPropertiesMeta(enum.EnumMeta):
         for val in cls:
             add_coerce_type(type(val.value))
 
-        for prop in cls.enum_properties:
-            values = classdict._ep_properties_[prop]
-            setattr(
-                cls,
-                f'_value2{prop}_map_',
-                dict(zip(cls._value2member_map_, values))
-            )
-            setattr(
-                cls,
-                prop,
-                property(
-                    lambda self, prop=prop:
-                    getattr(cls, f'_value2{prop}_map_').get(self.value)
+        # set properties onto the members
+        for idx, member in enumerate(cls.__members__.values()):
+            for prop, values in classdict._ep_properties_.items():
+                setattr(
+                    member,
+                    prop,
+                    values[idx]
                 )
-            )
 
         # we reverse to maintain precedence order for symmetric lookups
         for prop in reversed([
-            prop for prop in cls.enum_properties if prop.symmetric
+            prop for prop in cls._properties_ if prop.symmetric
         ]):
             for idx, val in enumerate(
                 reversed(classdict._ep_properties_[prop])
@@ -580,6 +626,24 @@ class FlagProperties(
     A Flag that supports properties
     """
 
+    def _generate_next_value_(
+            name,
+            start,
+            count,
+            last_values
+    ):  # pylint: disable=E0213
+        """
+        Intermixed property tuples can corrupt the last_values list with
+        tuples. This method ensures only ints are present in last_values and
+        delegates to the super class.
+        """
+        return enum.Flag._generate_next_value_(
+            name,
+            start,
+            count,
+            [val[0] if isinstance(val, tuple) else val for val in last_values]
+        )
+
 
 class IntFlagProperties(
     DecomposeMixin,
@@ -590,3 +654,21 @@ class IntFlagProperties(
     """
     An IntFlag that supports properties.
     """
+
+    def _generate_next_value_(
+            name,
+            start,
+            count,
+            last_values
+    ):  # pylint: disable=E0213
+        """
+        Intermixed property tuples can corrupt the last_values list with
+        tuples. This method ensures only ints are present in last_values and
+        delegates to the super class.
+        """
+        return enum.IntFlag._generate_next_value_(
+            name,
+            start,
+            count,
+            [val[0] if isinstance(val, tuple) else val for val in last_values]
+        )
